@@ -148,6 +148,10 @@ unsigned long uart1_baud = 38400;
 unsigned long uart2_baud = 38400;
 int mids[8] = {0, 1, 2, 3, 4, 5, 6, 7};
 
+#ifdef HASXML
+#include HASXML
+#endif
+
 struct propa_ptr props_cpu0[] = {
 	PROPA_PTR("device_type", "cpu", 4),
 	PROPA_PTR("name", "cpu0", 5),
@@ -278,6 +282,13 @@ void *prom_malloc(int size)
 	return (void *)prom_malloc_end;
 }
 
+void *prom_calloc(int size)
+{
+	void *r = prom_malloc(size);
+	_memset(r, 0, size);
+	return r;
+}
+
 static void leon_reboot(char *bcommand)
 {
 	outstr("PROM: Reboot  (not implemented)\n");
@@ -369,6 +380,12 @@ void outstr(char *str)
 	}
 }
 
+struct node *opprom_get_ambapp0(void) 
+{
+	return &nodes_lvl2[NODE_AMBAPP_IDX];
+}
+
+
 /* PROM Initialization routine, called from STARTUP when previous stages has
  * been successfully initialized
  */
@@ -393,9 +410,12 @@ void opprom_init(void)
 #endif
 	nodes_lvl2[NODE_CPU1_IDX - 1 + (ncpu - 1)].sibling = NULL;
 
-	/* Link in AMBA nodes into node tree */
+#ifdef CONFIG_CUSTOM_NODES
+	/* Link in AMBA nodes into node tree,
+	 * startup_ambapp.c insert childs using opprom_get_ambapp0() */
 	nodes_lvl2[NODE_AMBAPP_IDX].child = pnodes_amba;
-
+#endif
+	
 	/* Calculate IDPROM Checksum */
 	ptr = (unsigned char *)&idoprom;
 	for (i = cksum = 0; i <= 0x0E; i++)
