@@ -8,6 +8,7 @@
 #include <gptimer.h>
 #include <irqmp.h>
 #include <apbuart.h>
+#include <iommu.h>
 
 #ifdef CONFIG_AMPOPTS
 struct amp_opt {
@@ -18,6 +19,8 @@ struct amp_opt {
 };
 #endif
 
+void _memcpy4_pa(void *dst, void *src, int len);
+
 ambapp_dev *ambapp_get_first_dev(ambapp_core *core)
 {
 	if (core->ahbmst)
@@ -270,6 +273,107 @@ struct ambapp_bridge_driver {
 		ambapp_ahbdev *bridge);
 };
 
+/* proc.c */
+extern unsigned long amba_systemid;
+extern unsigned long ipi_num;
+struct propa_ptr props_ambapp_templ[] = {
+	PROPA_PTR("device_type", "ambapp", 7),
+	PROPA_PTR("name", "ambapp0", 8),
+	PROPA_PTR("systemid", &amba_systemid, 4),
+	PROPA_PTR("ioarea", 0, 4),
+	PROPA_PTR_END("ipi_num", &ipi_num, 4),
+};
+#define NAMEIDX   1
+#define IOAREAIDX 3
+
+struct node *ahb2ahb_create_bus_node(struct ambapp_bus *ainfo, unsigned long ioarea, int bus_idx) 
+{
+	struct node *n = 0;
+#ifndef FLATOFTREE
+	struct propa_ptr *pa;
+
+	/* create a bus node */
+	n = (struct node *)prom_calloc(sizeof(struct node));
+	/* initialize properties from template */
+	pa = (struct propa_ptr *)prom_calloc(sizeof(props_ambapp_templ));
+	_memcpy((void*)pa, (void*)&props_ambapp_templ, sizeof(props_ambapp_templ));
+	n->props = (struct prop *)pa;
+	/* customize ioarea */
+	pa[IOAREAIDX].value = prom_calloc(sizeof(int));
+	*(int *)pa[IOAREAIDX].value = ioarea;
+	/* customize name */
+	pa[NAMEIDX].value = prom_calloc(8);
+	_memcpy((void*)pa[NAMEIDX].value,"ambapp",6);
+	((char *)pa[NAMEIDX].value)[6] = '0' + bus_idx;
+	ainfo->ahbsn[bus_idx] = n;
+	ainfo->ahbsnn[bus_idx] = &n->child;
+#endif
+	return n;
+}
+
+/* AHB-AHB IOMMU Creation */
+void ahb2ahb_bus_create_iommu(struct ambapp_bus *ainfo, struct ahb_bus *parent,
+			ambapp_ahbdev *bridge)
+{
+	struct ahb_bus *subbus;
+	struct node *n;
+	struct iommu_regs *regs = 0;
+	ambapp_ahbdev *pdev;
+	int j;
+	
+	/* Add new bus to AMBA Information */
+	subbus = &ainfo->ahbs[ainfo->ahb_buses];
+	subbus->bus_idx = ainfo->ahb_buses++;
+	subbus->parent_bus_idx = parent->bus_idx;
+
+	/* save the bus-idx the io-mmu should node-link */
+	bridge->extra = subbus->bus_idx;
+	
+	/* Get info from Bridge's PnP custom information */
+	subbus->ioarea = -2; /* mark as ioarea, already scanned */
+	subbus->ffact = 1;
+	subbus->dir = 1; /* up */
+	subbus->isiommu = 1;
+	
+	/* get per bus node */
+	n = ahb2ahb_create_bus_node(ainfo, 0, subbus->bus_idx);
+
+	/* scan for reg address */
+	for (j = 0; j < 4; j++) {
+		if (bridge->address[j] && bridge->type[j] == AMBA_TYPE_AHBIO) {
+			regs = (struct iommu_regs *)bridge->address[j];
+			break;
+		}
+	}
+
+	/* scan the pnp information */
+	subbus->mst_cnt = 0;
+	if (regs) {
+		subbus->msts = (ambapp_ahbdev *)startup_malloc(16 * sizeof(ambapp_ahbdev));
+		pdev = &subbus->msts[0];
+		for (j = 0; j < 16; j++) {
+			uint32_t mst;
+			_memcpy4_pa((void *)&mst, (void *)&regs->mstconf[j], 4);
+			if (mst) {
+				struct ambapp_pnp_ahb ahb_buf;
+				_memset((void*)&ahb_buf, 0, sizeof(ahb_buf));
+				*(int *)&(ahb_buf.id) = mst;
+				*(int *)&(ahb_buf.mbar[0]) = 0; /* mst */
+				
+				/* Build a AHB Master Device */
+				pdev->dev_type = DEV_AHB_MST;
+				ambapp_dev_init_ahb(subbus, &ahb_buf, pdev);
+				
+				if (ambapp_fixup_dev((ambapp_dev *)pdev) == 0) {
+					subbus->mst_cnt++;
+					pdev++;
+				} 
+			}
+                }
+	}
+	
+}
+
 /* AHB-AHB Bridge Creation */
 void ahb2ahb_bus_create(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 			ambapp_ahbdev *bridge)
@@ -277,7 +381,8 @@ void ahb2ahb_bus_create(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 	struct ahb_bus *subbus;
 	unsigned long ioarea;
 	int i;
-
+	struct node *n;
+	
 	/* Is there a PnP area behind the bridge?*/
 	ioarea = bridge->userdef[1];
 	if (ioarea == 0)
@@ -288,15 +393,23 @@ void ahb2ahb_bus_create(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 		if (ainfo->ahbs[i].ioarea == ioarea)
 			return;
 
+	
 	/* Add new bus to AMBA Information */
 	subbus = &ainfo->ahbs[ainfo->ahb_buses];
 	subbus->bus_idx = ainfo->ahb_buses++;
 	subbus->parent_bus_idx = parent->bus_idx;
 
+	/* save the bus-idx the bridge should node-link */
+	bridge->extra = subbus->bus_idx;
+
 	/* Get info from Bridge's PnP custom information */
 	subbus->ioarea = ioarea;
 	subbus->ffact = (bridge->userdef[0] & AMBAPP_FLAG_FFACT) >> 4;
 	subbus->dir = (bridge->userdef[0] & AMBAPP_FLAG_FFACT_DIR);
+
+	/* get per bus node */
+	n = ahb2ahb_create_bus_node(ainfo, ioarea, subbus->bus_idx);
+	
 }
 
 /* AHB->APB Bridge */
@@ -308,7 +421,8 @@ void ahb2apb_bus_create(struct ambapp_bus *ainfo, struct ahb_bus *parent,
 	subbus = &ainfo->apbs[ainfo->apb_buses];
 	subbus->bus_idx = ainfo->apb_buses++;
 	subbus->parent_bus_idx = parent->bus_idx;
-
+	subbus->ahb_bus_idx = bridge->bus_idx;
+	
 	/* Get info from Bridge's PnP custom information */
 	subbus->ioarea = bridge->address[0];
 }
@@ -336,7 +450,7 @@ struct ambapp_bridge_driver bridgedrvs[] = {
 		.ver_min = 0,
 		.ver_max = 0xff,
 		.type = 0,
-		.create = ahb2ahb_bus_create
+		.create = ahb2ahb_bus_create_iommu
 	},
 	{
 		.vendor = VENDOR_GAISLER,
@@ -404,6 +518,7 @@ void ambapp_dev_init_ahb(
 	dev->device = ambapp_pnp_device(ahb->id);
 	dev->ver = ambapp_pnp_ver(ahb->id);
 	_memset(&dev->irq[0], 0, sizeof(dev->irq));
+	dev->extra = 0;
 	dev->irq[0] = ambapp_pnp_irq(ahb->id);
 	dev->bus_idx = bus->bus_idx;
 	dev->userdef[0] = (unsigned int)ahb->custom[0];
@@ -446,8 +561,10 @@ void ambapp_dev_init_apb(
 	dev->device = ambapp_pnp_device(apb->id);
 	dev->ver = ambapp_pnp_ver(apb->id);
 	_memset(&dev->irq[0], 0, sizeof(dev->irq));
+	dev->extra = 0;
 	dev->irq[0] = ambapp_pnp_irq(apb->id);
 	dev->bus_idx = bus->bus_idx;
+	dev->extra = bus->ahb_bus_idx;
 	dev->address = ambapp_pnp_apb_start(apb->iobar, bus->ioarea);
 	dev->mask = ambapp_pnp_apb_mask(apb->iobar);
 }
@@ -605,6 +722,8 @@ void ambapp_bus_init(
 	ainfo->ahb_buses = 1;
 
 	for (i = 0; i < AMBA_AHB_MAX; i++) {
+		if (ainfo->ahbs[i].ioarea == -2) /* iommu */
+			continue;
 		if (ainfo->ahbs[i].ioarea == 0)
 			break;
 		ambapp_scan_bus_ahb(ainfo, &ainfo->ahbs[i]);
@@ -1288,6 +1407,35 @@ int (*ambapp_prop_creators[])(ambapp_core *core, struct prop_std *buf) = {
 };
 #define PROP_CREATORS_NUM (sizeof(ambapp_prop_creators)/sizeof(void (*)(void)))
 
+#ifdef HASXML
+extern struct mksel_def sel[];
+extern struct mknod_def nod[];
+extern int selcnt;
+int
+match_node(ambapp_core *core, struct mksel_def **d, struct mksel_def *e, int *p) {
+	int r = 0;
+	struct mksel_def *s = *d;
+	ambapp_dev *a = ambapp_get_first_dev(core);
+    
+	while (s < e) {
+		if (s->ven && s->dev) {
+			if ((s->idx == -1 || s->idx == core->scanidx)
+			    && s->ven == a->vendor
+			    && s->dev == a->device) {
+				break;
+			}
+		} else if (core->index == s->pidx) {
+			break;
+		}
+		s++; if (p) *p = *p + 1;
+	}
+	if (s < e)
+		r = s->to - s->from;
+	*d = s;
+	return r;
+}
+#endif
+
 /* Add a AMBA Core to the end of the node tree */
 struct node *ambapp_node_create(ambapp_core *core)
 {
@@ -1296,7 +1444,7 @@ struct node *ambapp_node_create(ambapp_core *core)
 	unsigned int prop_buf[32];
 	struct prop_std *prop, *buf = (struct prop_std *)prop_buf;
 
-	n = prom_malloc(sizeof(struct node));
+	n = (struct node *)prom_malloc(sizeof(struct node));
 	n->child = NULL;
 	n->sibling = NULL;
 	n->props = NULL;
@@ -1311,6 +1459,44 @@ struct node *ambapp_node_create(ambapp_core *core)
 		prop_add(n, (struct prop *)prop);
 	}
 
+#ifdef HASXML
+	{
+		/* add extra properties|nodes from -xml <xmlfile> passed properties */
+		int sc, k, pit;
+		struct node *cn;
+		struct mksel_def *s = sel, *e = &sel[selcnt];
+		while(s < e && (sc = match_node(core, &s, e, 0))) {
+			if (!(s->isaddnode)) { /* add property */
+				struct mknod_def *nd = nod; struct propa_ptr *p;
+				for (k = s->from; k < s->to; k++) {
+					p = nd[k].p;
+					for (pit = 0; pit < nd[k].len; pit++) {
+						prop = (struct prop_std *)prom_malloc(sizeof(struct prop_std));
+						prop->options = PO_PTR | PO_NEXT;
+						prop->length = p[pit].length;
+						prop->name = p[pit].name;
+						prop->next = NULL;
+						prop->v.value = p[pit].value;
+						prop_add(n, (struct prop *)prop);
+					}
+				}
+			} else { /* add node-tree */
+				struct mknod_def *nd = nod;
+				for (k = s->from; k < s->to; k++) {
+					cn = (struct node *)prom_calloc(sizeof(struct node));
+					struct node **np = &n->child;
+					while (*np)
+						np = &((*np)->sibling);
+					*np = cn;
+					cn->child = nd[k].n.child;
+					cn->props = nd[k].n.props;
+				}
+			}
+			s++;
+		}
+	}
+#endif
+	
 	if (prop)
 		prop->options |= PO_END;
 
@@ -1318,25 +1504,82 @@ struct node *ambapp_node_create(ambapp_core *core)
 }
 
 struct node *pnodes_amba = NULL;
-struct node *pnodes_amba_last = NULL;
 
-void ambapp_nodes_init(void)
+void ambapp_nodes_init(struct node *b0)
 {
-	struct node *n;
+	struct node *n; int busidx = 0;
 	ambapp_core *core = ambapp_plb.cores;
+	
+#ifdef HASXML
+	ambapp_core *cn;
+	/* init ambapp_core.scanidx dublicates count */
+	while (core) {
+		ambapp_dev *a = ambapp_get_first_dev(core);
+		cn = core->next;
+		while (cn) {
+			ambapp_dev *b = ambapp_get_first_dev(cn);
+			if (a->device == b->device &&
+			    a->vendor == b->vendor) {
+				cn->scanidx++;
+			}
+			cn = cn->next;
+		}
+		core = core->next;
+	}
+#endif
+	/* ambapp0 */
+	ambapp_plb.ahbsn[0] = b0;
+	ambapp_plb.ahbsnn[0] = &b0->child;
 
+	core = ambapp_plb.cores;
 	while (core) {
 		n = ambapp_node_create(core);
 
 		if (ambapp_node_fixup(core, n) == 0) {
-			/* Insert Node last in chain */
-			if (pnodes_amba == NULL) {
-				pnodes_amba = n;
-				pnodes_amba_last = pnodes_amba;
-			} else {
-				pnodes_amba_last->sibling = n;
-				pnodes_amba_last = n;
+
+#ifndef FLATOFTREE
+			static struct ambapp_bridge_driver *b;
+			ambapp_dev *a = ambapp_get_first_dev(core);
+		
+			/* add it to the bus it belongs to */
+			busidx = a->bus_idx;
+			if (!ambapp_plb.ahbsn[busidx]) {
+				busidx = 0;
 			}
+			
+			switch (a->dev_type) {
+			case DEV_APB_SLV:
+				if (a->dev_type == DEV_APB_SLV) {
+					/* use the ahb-bus-idx that the AHB-APB bridge is connected to */
+					if (!ambapp_plb.ahbsn[busidx = a->extra])
+						busidx = 0;
+				}
+				break;
+			case DEV_AHB_MST:
+			case DEV_AHB_SLV:
+				if (core->ahbslv
+				    && (!(a->vendor == VENDOR_GAISLER && a->device == GAISLER_APBMST))
+				    && (b = ambapp_bridge_detect((ambapp_ahbdev *)a))) {
+					/* bridges connect with their slave */
+					a = (ambapp_dev *)core->ahbslv;
+					busidx = a->bus_idx;
+					/* dev.extra is the bridge to be child */
+					if (busidx < a->extra) {
+						struct node **np = &n->child;
+						while (*np)
+							np = &((*np)->sibling);
+						*np = ambapp_plb.ahbsn[a->extra];
+					}
+				} 
+				break;
+			}
+#else
+			busidx = 0;
+#endif
+
+			/* Insert Node last in chain */
+			*ambapp_plb.ahbsnn[busidx] = n;
+			ambapp_plb.ahbsnn[busidx] = &n->sibling;
 		}
 
 		core = core->next;
